import javax.swing.JPanel;
import data.DataRepresentation;
import java.awt.*;
import java.awt.image.BufferedImage;
import java.awt.geom.*;
import java.awt.event.*;

import drawing.*;


public class Plotter extends JPanel implements MouseListener, MouseMotionListener, IConstants{

  //Daten
  private DataRepresentation datrep;
  private double maxvalue;
  private Point[][] controlpoints;        // Kontrollpunkte des ThemeRiver
  private Point[][] controlpointshisto;   // Kontrollpunkte der Histogrammdarstellung

  //ThemeRiverDrawer
  private ThemeRiverDrawer drawer;
  private int interpstyle, reprstyle, pointstyle;
  private ColorList cl;

  //MagicLens
  private ThemeRiverDrawer lensdrawer;
  private BufferedImage lensimg;
  private Graphics2D lensgraphics;
  private int lensinterpstyle, lensreprstyle;
  private int lenspos, lenswidth;
  private boolean magiclens, resizeleft, resizeright, drag;
  private int resizestartx, dragstartx;



  //Zeichnen
  private boolean firsttime;
  private BufferedImage bi;
  private Graphics2D big;
  private Rectangle area;

  //Plottergre
  int width, height;

  //Konstanten fr die Interpolationsart
  public final static int LINEAR_INTERPOLATION = 0;
  public final static int NORMAL_SPLINE_INTERPOLATION = 1;
  public final static int ZEROSLOPE_SPLINE_INTERPOLATION = 2;
  public final static int MEANSLOPE_SPLINE_INTERPOLATION = 3;



  //Konstanten fr die Schriftarten
  private final static Font TITLEFONT = new Font("Arial", Font.BOLD, 16);
  private final static Font NORMALFONT = new Font("Arial", Font.PLAIN, 12);
  private final static Font TIMEDOMAINFONT = new Font("Arial", Font.BOLD, 12);
  private final static Font DATAFONT= new Font("Arial", Font.PLAIN, 10);

  //Konstanten fr die Darstellungsart
  public final static int RIVER_REPRESENTATION = 0;
  public final static int HISTO_REPRESENTATION = 1;

  //Konstanten fr das Zeichnen der Datenpunkte
  public final static int NO_POINTS = 0;
  public final static int SIMPLE_POINTS = 1;
  public final static int DATA_POINTS = 2;



  //Konstruktor fr den Plotter der Anwendung
  public Plotter(int width, int height, ColorList cl) {
    setSize(width, height);
    this.width = width;
    this.height = height;
    addMouseListener(this);
    addMouseMotionListener(this);

    this.cl = cl;
    interpstyle = LINEAR_INTERPOLATION;
    reprstyle = RIVER_REPRESENTATION;
    pointstyle = SIMPLE_POINTS;

    lensinterpstyle = MEANSLOPE_SPLINE_INTERPOLATION;
    lensreprstyle = HISTO_REPRESENTATION;
    lenspos = width/2;
    lenswidth = 50;
    lensimg = new BufferedImage(width, height, BufferedImage.TYPE_INT_RGB);
    lensgraphics = lensimg.createGraphics();
    lensgraphics.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
    lensgraphics.setFont(DATAFONT);
    magiclens = false;
    resizeleft = false;
    resizeright = false;
    drag = false;

    datrep = null;
    maxvalue = -1;
    controlpoints = null;
    firsttime = true;
    repaint();
  }



  //Konstruktor fr den Plotter fr den Export von Bildern
  public Plotter(Plotter oldplot, int width, int height) {
    setSize(width, height);
    this.width = width;
    this.height = height;
    addMouseListener(this);
    addMouseMotionListener(this);

    this.cl = oldplot.getColorList();
    interpstyle = oldplot.getInterpolationStyle();
    reprstyle = oldplot.getRepresentationStyle();
    pointstyle = oldplot.getPointStyle();



    int[] lenssettings = oldplot.getMagicLensSettings();
    lensinterpstyle = lenssettings[0];
    lensreprstyle = lenssettings[1];
    lenspos = (int)((double)lenssettings[2]*((double)width/(double)oldplot.getWidth()));
    lenswidth = (int)((double)lenssettings[3]*((double)width/(double)oldplot.getWidth()));
    lensimg = new BufferedImage(width, height, BufferedImage.TYPE_INT_RGB);
    lensgraphics = lensimg.createGraphics();
    lensgraphics.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
    lensgraphics.setFont(DATAFONT);
    magiclens = oldplot.existsMagicLens();



    datrep = null;
    maxvalue = -1;
    controlpoints = null;
    firsttime = false;


    setDataRepresentation(oldplot.getDataRepresentation());

    area = new Rectangle(width, height);
    bi = new BufferedImage(width, height, BufferedImage.TYPE_INT_RGB);
    big = bi.createGraphics();
    big.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
    update(big);
  }


  //setzt neuen Datensatz
  public void setDataRepresentation(DataRepresentation datrep) {
    if (datrep != null) {
      this.datrep = datrep;
      maxvalue = -1;
      controlpoints = null;
      calculateControlPoints();
      drawer = initializeThemeRiverDrawer(reprstyle, interpstyle);
      lensdrawer = initializeThemeRiverDrawer(lensreprstyle, lensinterpstyle);
    }
    repaint();
  }

  //setzt die Interpolationsart
  public void setInterpolationStyle (int style) {
    interpstyle = style;
    drawer = initializeThemeRiverDrawer(reprstyle, interpstyle);
    repaint();
  }

  //setzt die Darstellungsart
  public void setRepresentationStyle (int style) {
    reprstyle = style;
    drawer = initializeThemeRiverDrawer(reprstyle, interpstyle);
    repaint();
  }

  //setzt den Zeichenstil der Datenpunkte
  public void setPointStyle (int style) {
    pointstyle = style;
    repaint();
  }

  public void setMagicLens() {
    magiclens = true;
    lensdrawer = initializeThemeRiverDrawer(lensreprstyle, lensinterpstyle);
    //drawer.setMagicLens(initializeThemeRiverDrawer(reprstyle, interpstyle));
    repaint();
  }

  public void deleteMagicLens() {
    magiclens = false;
  }

  //setzt die Interpolationsart der MagicLens
  public void setMagicLensInterpolationStyle (int style) {
    lensinterpstyle = style;
    lensdrawer = initializeThemeRiverDrawer(lensreprstyle, lensinterpstyle);
    repaint();
  }

  //setzt die Darstellungsart der MagicLens
  public void setMagicLensRepresentationStyle (int style) {
    lensreprstyle = style;
    lensdrawer = initializeThemeRiverDrawer(lensreprstyle, lensinterpstyle);
    repaint();
  }


  public void setColorList(ColorList cl) {
    this.cl = cl;
    drawer = initializeThemeRiverDrawer(reprstyle, interpstyle);
    lensdrawer = initializeThemeRiverDrawer(lensreprstyle, lensinterpstyle);
    repaint();
  }


  public BufferedImage getImage() throws Exception {
    return bi;
  }

  public int getInterpolationStyle() {
    return interpstyle;
  }


  public int getRepresentationStyle() {
    return reprstyle;
  }

  public int getPointStyle() {
    return pointstyle;
  }


  public boolean existsMagicLens() {
    return magiclens;
  }

  public int[] getMagicLensSettings() {
    int[] settings = new int[4];
    settings[0] = lensinterpstyle;
    settings[1] = lensreprstyle;
    settings[2] = lenspos;
    settings[3] = lenswidth;

    return settings;
  }


  public DataRepresentation getDataRepresentation() {
    return datrep;
  }

  public ColorList getColorList() {
    return cl;
  }







  //Berechnet die Kontrollpunkte durch die die Splines gehen mssen
  private void calculateControlPoints() {

    maxvalue = datrep.getMaximumSumValue();

    //Erstellen der Datenstruktur
    int datanumber = datrep.countData();
    int colnumber = (datrep.getData(0)).length;

    controlpoints = new Point[datanumber][colnumber+1];
    controlpointshisto = new Point[datanumber][colnumber+1];


    //Ratio: Wieviele Pixel pro Einheit
    int h = height-MARGIN;

    double verticalratio = (double)h/maxvalue;
    double horizontalratio = (double)width/(double)datanumber;


    for (int i = 0; i < datanumber; i++) {
      double[] currdata = datrep.getData(i);

      double sum = datrep.getRowSum(i);

      // ThemeRiver Kontrollpunkte
      int offset = (int)(MARGIN/2 + (maxvalue-sum)*0.5*verticalratio);

      for (int j = 0; j < colnumber; j++) {
        controlpoints[i][j] = new Point((int)((i+0.5)*horizontalratio), offset);
        offset += (int)(currdata[j]*verticalratio);
      }
      //Abschlieende Punktreihe, die den Themeriver nach unten begrenzt.
      controlpoints[i][colnumber] = new Point((int)((i+0.5)*horizontalratio), offset);


      // Histogramm Kontrollpunkte
      int histooffset = (int)(MARGIN / 2);

      for (int j = 0; j < colnumber; j++) {
          controlpointshisto[i][j] = new Point((int)((i+0.5)*horizontalratio), histooffset);
          histooffset += (int)(((double)currdata[j] / sum) * (double)(height-MARGIN));
      }
      controlpointshisto[i][colnumber] = new Point((int)((i+0.5)*horizontalratio), height - (int)((MARGIN / 2)));
    }


  }


  //Initialisiert den ThemeRiverDrawer gem dem DrawingStyle
  private ThemeRiverDrawer initializeThemeRiverDrawer(int reprstyle, int interpstyle) {
    ThemeRiverDrawer newdrawer;
    switch (interpstyle) {
      case NORMAL_SPLINE_INTERPOLATION: {
          if (reprstyle == RIVER_REPRESENTATION) {
            newdrawer = new SplineThemeRiverDrawer(controlpoints, width, height, cl);
        }
        else {
           newdrawer = new SplineThemeRiverDrawer(controlpointshisto, width, height, cl);
        }
        break;
      }
            case ZEROSLOPE_SPLINE_INTERPOLATION: {
                if (reprstyle == RIVER_REPRESENTATION) {
            newdrawer = new SplineThemeRiverDrawer_ZeroSlope(controlpoints, width, height, cl);
        }
        else {
           newdrawer = new SplineThemeRiverDrawer_ZeroSlope(controlpointshisto, width, height, cl);
        }
        break;
      }
      case MEANSLOPE_SPLINE_INTERPOLATION: {
          if (reprstyle == RIVER_REPRESENTATION) {
            newdrawer = new SplineThemeRiverDrawer_MeanSlope(controlpoints, width, height, cl);
        }
        else {
           newdrawer = new SplineThemeRiverDrawer_MeanSlope(controlpointshisto, width, height, cl);
        }
        break;
      }
      default: {
          if (reprstyle == RIVER_REPRESENTATION) {
            newdrawer = new LinearThemeRiverDrawer(controlpoints, width, height, cl);
        }
        else {
           newdrawer = new LinearThemeRiverDrawer(controlpointshisto, width, height, cl);
        }
            break;
      }
    }
    return newdrawer;

  }



  //Paint-Methode
  public void paintComponent(Graphics g){
    super.paintComponent(g);
    update(g);
  }


  //Update-Methode beinhaltet Erstellung des Bildes
  public void update(Graphics g){
    Graphics2D g2 = (Graphics2D)g;

    if(firsttime){
      area = new Rectangle(width, height);
      bi = (BufferedImage)createImage(width, height);
      big = bi.createGraphics();
      big.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
      firsttime = false;
      System.out.println("neues Bild fertig");
    }

    // Clears the rectangle that was previously drawn.
    big.setColor(Color.WHITE);
    big.fillRect(0, 0, area.width, area.height);


    if (datrep != null) {
      //Reihenfolge muss eingehalten werden !!!
      drawer.draw(big);
      drawTimeLine(big);
      drawMagicLens(big);
      drawPoints(big);
    }


    // Draws the buffered image to the screen.
    g2.drawImage(bi, 0, 0, this);
  }




  //Zeichnet Zeitskala
  private void drawTimeLine(Graphics2D g2d) {

    //Lschen der zu beschriftenden Bereiche
    g2d.setColor(Color.WHITE);
    g2d.fillRect(0, 0, width, MARGIN/2);
    g2d.fillRect(0, height-(MARGIN/2), width, MARGIN/2);



    int number = datrep.countData();

    double ratio = (double)width/(double)number;

    g2d.setFont(NORMALFONT);
    FontMetrics metrics = big.getFontMetrics();


    g2d.setColor(Color.BLACK);
    for (int i = 0; i < number; i++) {
      g2d.drawLine((int)((i+1)*ratio), height, (int)((i+1)*ratio), height-40);


      String currstr = datrep.getTimeData(i);



      while((metrics.stringWidth(currstr) > (ratio-10)) && (currstr.length() > 3)) {
        currstr = currstr.substring(0, currstr.length()-3) + ".";
      }

      g2d.drawString(currstr, (int)((i)*ratio + (ratio-metrics.stringWidth(currstr))/2), height-30);
    }


    //Einheit der Zeitleiste
    g2d.setFont(TIMEDOMAINFONT);
    metrics = big.getFontMetrics();

    String timedomain = datrep.getTimeDomain();
    int stringwidth = metrics.stringWidth(timedomain);
    int xpos = (getWidth() - stringwidth)/2;

    g2d.setColor(Color.WHITE);
    g2d.fillRect(xpos-5, height-25, stringwidth+10, 20);

    g2d.setColor(Color.BLACK);
    g2d.drawString(timedomain, xpos, height-10);

  }




  //Zeichnet Datenpunkte ein
  private void drawPoints(Graphics2D g2d) {

    if (pointstyle != NO_POINTS) {
      g2d.setColor(Color.BLACK);
      g2d.setFont(DATAFONT);

      if (reprstyle == RIVER_REPRESENTATION) {
        for (int i = 0; i < controlpoints.length; i++) {
          for (int j = 0; j < controlpoints[0].length; j++) {

            drawPoint(controlpoints, i, j, g2d);

          }
        }
      }

      else {
        for (int i = 0; i < controlpointshisto.length; i++) {
          for (int j = 0; j < controlpointshisto[0].length; j++) {

            drawPoint(controlpointshisto, i, j, g2d);

          }
        }
      }

    }

  }


  //Zeichnet den Datenpunkt in dem gewnschten Format
  private void drawPoint(Point[][] pointlist, int index1, int index2, Graphics2D g2d) {



    Point currpoint = pointlist[index1][index2];

    if ((magiclens) && (isInMagicLens(currpoint.x, currpoint.y))) {
      return;
    }

    g2d.drawOval(currpoint.x-3, currpoint.y-3, 6, 6);

    if ((pointstyle == DATA_POINTS) && (index2 > 0)) {
      String strval;
      if (reprstyle == RIVER_REPRESENTATION) {
        double value = (datrep.getData(index1))[index2-1];
        strval = Double.toString(value);
      }
      else {
        int value = (datrep.getPercentageData(index1))[index2-1];
        strval = Integer.toString(value) + " %";
      }

      FontMetrics metrics = g2d.getFontMetrics();

      int ypos = (currpoint.y + pointlist[index1][index2-1].y)/2;

      g2d.drawString(strval, currpoint.x-(metrics.stringWidth(strval)/2), ypos);
    }


  }


  private void drawMagicLens(Graphics2D g2d) {

    if (magiclens) {

      lensgraphics.setColor(Color.WHITE);
      lensgraphics.fillRect(0, 0, width, height);
      lensdrawer.draw(lensgraphics);


      g2d.drawImage(lensimg, lenspos-lenswidth/2, MARGIN/2, lenspos+lenswidth/2, height-MARGIN/2,
      lenspos-lenswidth/2, MARGIN/2, lenspos+lenswidth/2, height-MARGIN/2, Color.WHITE, this);

      g2d.setColor(Color.BLACK);
      g2d.drawRect(lenspos-lenswidth/2, MARGIN/2, lenswidth, height-MARGIN);
      g2d.drawRect(lenspos-lenswidth/2+1, MARGIN/2+1, lenswidth, height-MARGIN-2);


    }
  }

  private boolean isInMagicLens(int x, int y) {
    if ((x >= (lenspos-lenswidth/2)) && (x <= (lenspos+lenswidth/2))) {
      if ((y > (MARGIN/2)) && (y <= height-(MARGIN/2))) {
        return true;
      }
    }
    return false;
  }


  private boolean isOnLeftMagicLensBorder(int x, int y) {
    if ((x >= (lenspos-lenswidth/2-3)) && (x <= (lenspos-lenswidth/2+3))) {
      if ((y > (MARGIN/2)) && (y <= height-(MARGIN/2))) {
        return true;
      }
    }

    return false;
  }



  private boolean isOnRightMagicLensBorder(int x, int y) {

    if ((x >= (lenspos+lenswidth/2-3)) && (x <= (lenspos+lenswidth/2+3))) {
      if ((y > (MARGIN/2)) && (y <= height-(MARGIN/2))) {
        return true;
      }
    }

    return false;
  }




  //MouseListener Methoden
  public void mouseClicked(MouseEvent e) {

  }
  public void mouseEntered(MouseEvent e) {}
  public void mouseExited(MouseEvent e) {}
  public void mousePressed(MouseEvent e) {



    if (magiclens) {

      if (isOnLeftMagicLensBorder(e.getX(), e.getY())) {
        resizeleft = true;
        setCursor(new Cursor(Cursor.E_RESIZE_CURSOR));
        resizestartx = e.getX();
        System.out.println("hilfe");
        return;
      }


      if (isOnRightMagicLensBorder(e.getX(), e.getY())) {
        resizeright = true;
        setCursor(new Cursor(Cursor.E_RESIZE_CURSOR));
        resizestartx = e.getX();
        System.out.println("hilfe");
        return;
      }


      if (isInMagicLens(e.getX(), e.getY())) {
        drag = true;
        dragstartx = e.getX();
      }
    }

  }
  public void mouseReleased(MouseEvent e) {
    if (drag) {
      drag = false;
    }
    if (resizeleft) {
      resizeleft = false;
    }
    if (resizeright) {
      resizeright = false;
    }
  }


  //MouseMotionListener Methoden
  public void mouseDragged(MouseEvent e) {

    if (resizeleft) {
      if ((lenswidth + 2*(resizestartx-e.getX())) > 10) {
        lenswidth += 2*(resizestartx-e.getX());
        resizestartx = e.getX();
        repaint();
      }
      else {
        lenswidth = 10;
        repaint();
      }
    }

    if (resizeright) {
      if ((lenswidth + 2*(e.getX()-resizestartx)) > 10) {
        lenswidth += 2*(e.getX()-resizestartx);
        resizestartx = e.getX();
        repaint();
      }
      else {
        lenswidth = 10;
        repaint();
      }
    }


    if (drag) {
      if (((lenspos + e.getX() - dragstartx) < width)  && ((lenspos + e.getX() - dragstartx) > 0)) {
        lenspos += e.getX()-dragstartx;
        dragstartx = e.getX();
        repaint();
      }
    }
  }



  public void mouseMoved(MouseEvent e) {
    if (magiclens) {
      if (isOnLeftMagicLensBorder(e.getX(), e.getY())) {
        setCursor(new Cursor(Cursor.E_RESIZE_CURSOR));
        return;
      }

      if (isOnRightMagicLensBorder(e.getX(), e.getY())) {
        setCursor(new Cursor(Cursor.E_RESIZE_CURSOR));
        return;
      }

      setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
    }



  }


}